   /* next-hack simple 16-bit audio player,  Copyright 2017 Nicola Wrachien and Franco Gobbo (next-hack.com)
   * It plays an audio file (audio.raw) from the root of the SD card.
   * 
   *  This file is part of next-hack's next-hack simple 16-bit audio player.
  
   *  This program is free software: you can redistribute it and/or modify
   *  it under the terms of the GNU General Public License as published by
   *  the Free Software Foundation, either version 3 of the License, or
   *  (at your option) any later version.
  
   *  This program  is distributed in the hope that it will be useful,
   *  but WITHOUT ANY WARRANTY; without even the implied warranty of
   *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   *  GNU General Public License for more details.
  
   *  You should have received a copy of the GNU General Public License
   *  along with next-hack's sd-test.  If not, see <http://www.gnu.org/licenses/>.   
   * It plays a raw audio-video file (named audiovid.raw), generated by the bmp2raw utility.
   * 
   * This program assumes your SD can operate at 8MHz (though you can modify mmc_avr_spi.c in order to use slower clock speeds).
   * 
   * This program is compatible with the following file format:
   * file format: raw
   * rate: the rate you set using SAMPLE_RATE
   * number of bits: 16
   * sample format: signed-integer (if unsigned, replace "OCR1BL = 128 + *p++;" with  "OCR1BL = *p++;")  
   * 
   * With almost no effort you can make it capable of playing regular wav files!
   * You'll need to apply minor changes on the ISR and to skip the initial wav header
   * 
   * Enjoy!
   */
#include "ff.h"
//#define USETIMER1PWM              // uncomment this if you are using timer 1 as pwm output! This will give you better quality!
#define SAMPLE_RATE  20000UL          // in Hz. You can use up to 31373, just be sure to use the correct sample!
#define SAMPLES_PER_BANK 0U                     // 0 = 256 16-bit valid samples in current bank. This is used with video playback, because not every sound sector will have 256 valid samples.

#ifndef F_CPU
  #define F_CPU 16000000UL
#endif  
//

FATFS g_fs;            
FIL g_file;   
volatile unsigned char g_currentBankNumber;     // current audio bank number being played. Used to understand to which bank we should read/write
volatile char* g_pCurrentSample;                // pointer to current sample
volatile unsigned char  g_currentSample;        // current samples that will be outputted. Used to avoid 16-bit operations.
char audioBank1[512], audioBank2[512];          // 2 512-bytes wide buffers allows glitch-free playback, without much headache concerning synchronizations. You could also avoid double buffering, freeing 512 bytes of ram.
volatile char g_readRequest;                    // it's set to 1 when we finished outputting a buffer and we cann then fill it with fresh data.
//
void enableAudioTimer(void);
void initAudioTimer(void);

void setup()
{
  unsigned char res;
  DDRC |= 0x3 << 4;  // needed to deselect display while configuring sd
  PORTC |= 0x3 << 4; //  needed to deselect display while configuring sd
  pinMode(LED_BUILTIN, OUTPUT);
  res = f_mount(&g_fs, "0:", 1);
  if (res)
  {
    digitalWrite(LED_BUILTIN, HIGH);   // turn the LED on (HIGH is the voltage level)
    while(1);   // cannot mount the SD: our journey ends here!
  }
  cli(); 
  initAudioTimer();  
  sei();
}
void loop()
{
  unsigned int bytesAvailable;
  uint8_t res;
  // Note: this program with minor modifications can be changed to play actual wav files! 
  res = f_open(&g_file, "audio.raw",FA_READ);
  
  if (res)
  {
    digitalWrite(LED_BUILTIN, HIGH);   // turn the LED on (HIGH is the voltage level)
    while(1);   // Cannot open the file: our journey ends here!
  }

  // now that our audio file is open, fill up the first buffer, and initialize the variables accordingly.
  g_currentBankNumber = 0;         // we initially play from bank 1
  g_pCurrentSample = audioBank1;                    
  g_currentSample = 0;   
  g_readRequest = 1;           // we already have to fill bank2, while bank1 is playing.
  //
  f_read (&g_file, audioBank1 , 512, &bytesAvailable);  // fill audio bank1
  enableAudioTimer();
  //

  while (1)
  {   
    if (g_readRequest)    // have we finished outputting all the samples ?
    {
      g_readRequest = 0;
      if (g_currentBankNumber)
      {
        f_read (&g_file, audioBank1 , 512, &bytesAvailable);  // fill audio bank 1
      }
      else
      {
        f_read (&g_file, audioBank2 , 512, &bytesAvailable);  // fill audio bank 2
      }
      if (bytesAvailable == 0)
      {
        // we reached the EOF. return.
        return;
      }             
    }
  }  
}
void enableAudioTimer(void)
{
  #ifndef USETIMER1PWM
    TCCR0B |= (1 << CS00);
    TCCR1B |= (1 << CS10);
  #else
    TCCR1B = (1 << WGM13) | (0 << WGM12) | (0 << CS12) | (0 << CS11) | (1 << CS10);
  #endif
}
void initAudioTimer(void)
{
  #ifndef USETIMER1PWM
    // Configure timer 0 for phase correct PWM
    TCCR0A = (1 << COM0A1) | (1 << COM0A0) | (1 << COM0B0) | (1 << COM0B1) | (1 << WGM00) | (0 << WGM01); // phase correct PWM.
    TCCR0B = (0 << WGM02) | (0 << CS02) | (0 << CS01) | (0 << CS00);     // Top is always 0xFF, no prescaling, timer initially disabled
    TCNT0 = 0;
    TIMSK0 = 0;       // disable all the intterupts on timer 0. This will also affect the delay/timekeeping functions, but we can live with it.
    DDRD |= (1 << PD5) | (1 << PD6);    // enable outputs 
    // Configure timer 1 at CTC, for sample rate generation.
    TCCR1A = 0;
    TCCR1B = (1 << WGM12);
    TCCR1C = 0;
    TIMSK1 = (1 << OCF1A);
    TCNT1 = 1;      
    OCR1A = (0.5 + F_CPU / SAMPLE_RATE) -1 ;      // count 0...509
  #else
    TIMSK0 = 0;       // disable all the intterupts on timer 0. This will also affect the delay/timekeeping functions, but we can live with it.
    //
    DDRB |= (1 << PB1) | (1 << PB2);    // enable PWM outputs (timer 1) 
    // Configure timer 1 for phase correct PWM. This is also the sample rate generator!
    TCCR1A = (1 << COM1A1) | (0 << COM1A0) | (1 << COM1B1) | (0 << COM1B0) | (0 << WGM11) | (0 << WGM10); // phase and frequency correct PWM.
    TCCR1B = (1 << WGM13) | (0 << WGM12) | (0 << CS12) | (0 << CS11) | (0 << CS10);     // Top is ICR1, no prescaling, timer initially disabled
    TCCR1C = 0;
    ICR1 = ((unsigned long) F_CPU / (2 * SAMPLE_RATE));    // SAMPLE_RATE = F_CPU / (2 * (ICR1 + 1 ) - 2)
    TIMSK1 = (1 << ICIE1);   // compare register set on top (ICR1) but updated at bottom (they are double buffered).
    TCNT1 = 0;
    OCR1A = 0;
    OCR1B = 0;
  #endif
}
#ifndef USETIMER1PWM
ISR(TIMER1_COMPA_vect) 
{
  // In this ISR we must be very quick! The C code has been optimized for speed (in particular to reduce accesses to volatile variables).
  // further optimizations could be achieved in ASM
  // The total ISR time is about 3.5us
  char *p = g_pCurrentSample;   // pointer to the current sample
  unsigned char currentSample = g_currentSample;  // we do this because g_currentSample is volatile and operating directly on it steals a lot of cycles
  OCR0A = *p++;      // LSB  
  OCR0B = 128 + *p++;        // MSB  
  if (currentSample == ((unsigned char) (SAMPLES_PER_BANK - 1) ) )    // in video mode we will use the full 512-bytes sector.
  {
    if (g_currentBankNumber)      // we don't do g_currentBankNumber = 1 - g_currentBankNumber as g_currentBankNumber is volatile and this would result in one more access
    {
      g_currentBankNumber = 0;
      g_pCurrentSample = audioBank1;
    }
    else
    {
      g_currentBankNumber = 1;
      g_pCurrentSample = audioBank2;
    }
    g_currentSample = 0;    
    g_readRequest = 1;      // issue a new read request.
  }
  else
  {
    g_pCurrentSample = p;           // because g_pCurrentSample is volatile, it should be faster incrementing p and then store back p to g_pCurrentSample
    g_currentSample = currentSample + 1;  
  }
}
#else
ISR(TIMER1_CAPT_vect) 
{
  // In this ISR we must be very quick! The C code has been optimized for speed (in particular to reduce accesses to volatile variables).
  // further optimizations could be achieved in ASM
  // The total ISR time is about 3.5us
  char *p = g_pCurrentSample;   // pointer to the current sample
  unsigned char currentSample = g_currentSample;  // we do this because g_currentSample is volatile and operating directly on it steals a lot of cycles
  OCR1AL = *p++;      // LSB  
  OCR1BL = 128 + *p++;        // MSB  
  if (currentSample == ((unsigned char) (SAMPLES_PER_BANK - 1) ) )    // in video mode we will use the full 512-bytes sector.
  {
    if (g_currentBankNumber)      // we don't do g_currentBankNumber = 1 - g_currentBankNumber as g_currentBankNumber is volatile and this would result in one more access
    {
      g_currentBankNumber = 0;
      g_pCurrentSample = audioBank1;
    }
    else
    {
      g_currentBankNumber = 1;
      g_pCurrentSample = audioBank2;
    }
    g_currentSample = 0;    
    g_readRequest = 1;      // issue a new read request.
  }
  else
  {
    g_pCurrentSample = p;           // because g_pCurrentSample is volatile, it should be faster incrementing p and then store back p to g_pCurrentSample
    g_currentSample = currentSample + 1;  
  }
}
#endif
